import pytest
from a_sync import cgather
from brownie import chain

from tests.fixtures import async_uni_v1, mutate_addresses
from y.networks import Network
from y.prices import magic
from y.prices.dex.uniswap import v3
from y.prices.dex.uniswap.uniswap import uniswap_multiplexer

V1_TOKENS = {
    Network.Mainnet: [
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
    ],
}.get(chain.id, [])

V2_TOKENS = {
    Network.Mainnet: [
        "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
        "0xBA11D00c5f74255f56a5E366F4F77f5A186d7f55",
        "0xc00e94Cb662C3520282E6f5717214004A7f26888",
        "0xD533a949740bb3306d119CC777fa900bA034cd52",
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "0x6810e776880C02933D47DB1b9fc05908e5386b96",
        "0xc944E90C64B2c07662A292be6244BDf05Cda44a7",
        "0x514910771AF9Ca656af840dff83E8264EcF986CA",
        "0x0F5D2fB29fb7d3CFeE444a200298f468908cC942",
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
        "0xec67005c4E498Ec7f55E092bd1d35cbC47C91892",
        "0x4fE83213D56308330EC302a8BD641f1d0113A4Cc",
        "0xC011a73ee8576Fb46F5E1c5751cA3B9Fe0af2a6F",
        "0x04Fa0d235C4abf4BcF4787aF4CF447DE572eF828",
        "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
        "0xdAC17F958D2ee523a2206206994597C13D831ec7",
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e",
        "0xE41d2489571d322189246DaFA5ebDe1F4699F498",
    ],
}.get(chain.id, [])

V2_TOKENS = mutate_addresses(V2_TOKENS)


@pytest.mark.parametrize("token", V1_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v1(token, async_uni_v1):
    price, alt_price = await cgather(
        async_uni_v1.get_price(token, None),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    # check if price is within 5% range
    assert price == pytest.approx(alt_price, rel=5e-2)


@pytest.mark.parametrize("token", V2_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v2(token):
    deepest_router = await uniswap_multiplexer.deepest_router(token, sync=False)
    price, alt_price = await cgather(
        deepest_router.get_price(token, skip_cache=True, sync=False),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    assert price == pytest.approx(alt_price, rel=5e-2)


@pytest.mark.parametrize("token", V2_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v3(token):
    price, alt_price = await cgather(
        v3.uniswap_v3.get_price(token, skip_cache=True, sync=False),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    assert price == pytest.approx(alt_price, rel=5e-2)



from decimal import Decimal
from typing import List, Dict

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, ContractBase
from y.contracts import Contract
from y.datatypes import Address, Block
from y.exceptions import UnsupportedNetwork
from y.networks import Network
from y.utils import gather2
from y.utils.cache import a_sync_ttl_cache

registry = "0xA50d4E7D8946a7c90652339CDBd262c375d54D99"


class DieselPool(ContractBase):
    @a_sync.aka.cached_property
    async def contract(self) -> Contract:
        return await Contract.coroutine(self.address)

    __contract__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.aka.cached_property
    async def diesel_token(self) -> ERC20:
        contract = await self.__contract__
        try:
            return ERC20(await contract.dieselToken, asynchronous=self.asynchronous)
        except (
            AttributeError
        ):  # NOTE: there could be better ways of doing this with hueristics, not sure yet
            return ERC20(self.address, asynchronous=self.asynchronous)

    __diesel_token__: HiddenMethodDescriptor[Self, ERC20]

    @a_sync.aka.cached_property
    async def underlying(self) -> ERC20:
        contract = await self.__contract__
        return ERC20(await contract.underlyingToken, asynchronous=self.asynchronous)

    __underlying__: HiddenMethodDescriptor[Self, ERC20]

    async def exchange_rate(self, block: Block) -> Decimal:
        underlying: ERC20
        pool, underlying = await gather2(self.__contract__, self.__underlying__)
        scale = await underlying.__scale__
        return Decimal(
            await pool.fromDiesel.coroutine(scale, block_identifier=block)
        ) / Decimal(scale)

    async def get_price(
        self, block: Block, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Decimal:
        underlying, exchange_rate = await gather2(
            self.__underlying__, self.exchange_rate(block, sync=False)
        )
        und_price = await underlying.price(block, skip_cache=skip_cache, sync=False)
        return Decimal(und_price) * exchange_rate


class Gearbox(a_sync.ASyncGenericBase):
    def __init__(self, *, asynchronous: bool = False):
        """Initialize a Gearbox instance.

        Raises:
            UnsupportedNetwork: If the current network is not Ethereum Mainnet.

        Args:
            asynchronous: Whether to use asynchronous operations. Defaults to False.

        Example:
            >>> from y.prices.gearbox import Gearbox
            >>> gearbox = Gearbox(asynchronous=True)
            >>> # This will raise an UnsupportedNetwork exception if not on Mainnet
        """
        if chain.id != Network.Mainnet:
            raise UnsupportedNetwork("gearbox not supported on this network")
        self.asynchronous = asynchronous
        self._dtokens: Dict[ERC20, DieselPool] = {}
        super().__init__()

    @a_sync.aka.cached_property
    async def registry(self) -> Contract:
        return await Contract.coroutine(registry)

    @a_sync_ttl_cache
    async def pools(self) -> List[DieselPool]:
        registry = await self.registry
        return [
            DieselPool(pool, asynchronous=self.asynchronous)
            for pool in await registry.getPools
        ]

    async def diesel_tokens(self) -> Dict[ERC20, DieselPool]:
        pools: List[DieselPool] = await self.pools(sync=False)
        return dict(
            zip(await DieselPool.diesel_token.map(pools).values(pop=True), pools)
        )

    async def is_diesel_token(self, token: Address) -> bool:
        return token in await self.diesel_tokens(sync=False)

    async def get_price(
        self, token: Address, block: Block, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Decimal:
        dtokens = await self.diesel_tokens()
        return await dtokens[token].get_price(block, skip_cache=skip_cache, sync=False)


try:
    gearbox = Gearbox(asynchronous=True)
except:
    gearbox = set()
